import os
import shutil
import weakref

from binascii import b2a_hex, a2b_hex
from os.path import join, exists, isdir, getmtime
from threading import RLock

from Interface import Base

from Products.Gideon import gideon_home
from Products.Gideon.Log  import LogFactory

log = LogFactory('TxnFs')

class TxnFsError(Exception): pass
class TxnFsEncodingError(TxnFsError): pass        
class TransactionRegistration: pass

#################################
# Some Util Functions
def copy_stream(instream, outstream, bufsize=8196):

    instream.seek(0,2)
    size = instream.tell()
    tsize = size
    instream.seek(0)

    while size:
        buf = instream.read(buffer_size)
        size -= buffer_size
        outstream.write(buf)

    return tsize

def decode_path(path):
    """
    decode an encoded path
    """

    return a2b_hex(path)

def encode_path(path):
    """
    encode a path in a string format that
    is an os acceptable directory format
    """

    if path.count('..'):
        raise TxnFsEncodingError("path contains illegal characters")

    return b2a_hex(path)
    
def file_collector(directory, res=[]):
    """ recursive file collector """

    res.extend( directory.get_files() )
    
    for d in directory.get_directories():
        
        res.extend( file_collector(d) )

    return res

#################################
              
class TransactionalFileSystemDirectory:

    _rm_flag = 'rm'

    def __init__(self, path, temppath, relative=1):
        self.root = path
        self.temp = temppath
        self._relative = 1
        
    def add_file(self, path, *args):

        log.debug('adding file %s'%path)
        
	path = self.resolve_path(path)
        
        name = encode_path(path)
        
        registry.add_file(name)
        
        rfh  = open(path)

        self.add_file_object(path, rfh, lock_file=0)

    def add_file_object(self, path, fileobj, lock_file=1):

        log.debug('adding file obj %s'%path)

	path = self.resolve_path(path)
        
        name = encode_path(path)
        
        if lock_file:
            registry.add_file(name)

        try:
            
            fh = open(join(self.tmpdir, name), 'w')
            copy_stream(fileobj, fh)

        except:
            registry.remove_file(name)
            raise

    def add_directory(self,  path, *args):

        log.debug('adding directory obj %s'%path)

	path = self.resolve_path(path)        

        name = encode_path(path)

        log.debug('adding directory to temp')
        print join(self.temp,name)
        
        os.mkdir( join(self.temp, name) )


    def create(self):
        """ try to create the path if it doesn't exist """

        if exists(self.root):
            return
        else:
            os.mkdir(self.root)

    def exists_p(self, path):

	path = self.resolve_path(path)

        return exists(path)

    def get_directory(self, path, relative=1):

	path = self.resolve_path(path)        

        if not exists(path):
            raise TxnFsError, "path does not exist"            
        elif not isdir(path):
            raise TxnFsError, "not a directory"            
        
        return self.__class__(path, self.temp, relative)
        
    def get_directories(self, relative=1):
        
        return [self.__class__(join(self.root, d), self.temp, relative)  \
                               for d in os.listdir(self.root) \
                               if isdir(join(self.root,d))]

    def get_file(self, path):

        path = self.resolve_path(path)

        assert exists(path) and not isdir(path)        

        ####
        return TxnFile(path, self)

    def get_files(self, recurse=0):

        files = []
        
        if recurse:
            for f in file_collector(self):
                files.append(TxnFile(f, self))                
        else:
            for f in os.listdir(self.root):
                if isdir(join(self.root, f)): continue
                files.append(TxnFile(f, self))
                
        return files
        
    def get_mod_time(self, path):

        return getmtime(path)

    def get_absolute_path(self):

        return self.root

    def is_relative(self):
        
        return self._relative

    # future as a callback from file when getting an output stream
    #def _modify_file(self, path):
        
    def remove_file(self, path, *args):

        path = self.resolve_path(path)

        registry.add_file(path)

        os.rename(path, self._rm_flag+encode_path(path))

    def remove_directory(self, path, force=0):

        directory = self.get_directory(path)

        files = file_collector( directory )

        if files and force:
            registry.add_files(map(lambda x: x.get_path(), files))
        elif files:
            raise TxnFsError, "directory not empty"

        path = self.resolve_path(path)

        os.rename(path, self._rm_flag+encode_path(path))        

    def resolve_path(self, path):

        if path.count('..'):
            raise TxnFsError, "relative addressing via .. not allowed"
        
        if self._relative:
            path = join(self.root, path)
        else:
            if not path.startswith(self.root):
                raise TxnFsError, "directory not under txn fs" 

        return path

    def set_path(self, path, create=0):

        if not self.exists_p(path) and create:
            os.mkdir(path)

        self.root = path

    def set_relative(self, relative):

        self._relative = int(relative)

class TxnFile:
    """
    does not exhibit any txn characterstics """

    def __init__(self, path, directory):
        self.path = path
        self.directory = weakref(directory)
        
    def size(self):

        return os.stat(join(self.directory().absolute_path(), self.path))[6]

    def input_stream(self):
        
        return open(join(self.directory().absolute_path(),self.path))

    def output_stream(self):
        raise "UnImplemented"


class PackageFileSystem(TransactionalFileSystemDirectory):

    def __getitem__(self, key):

        bucket = key[0]

        directory = self.get_directory(bucket)

        return directory.get_directory(key)

    def add_package_directory(self, pkgid):

        bucket = pkgid[0]

        directory = self.get_directory(bucket)

        directory.add_directory(pkgid)

_registration_storage = {}
class TxnFsTransactionRegistration:

    def __init__(self): pass

    def register(self, *args): pass
    
    def abort(self, *args): pass

    def commit(self, *args): pass

    def tpc_vote(self, *args): pass
    
    def tpc_finish(self, *args): pass

    def tpc_abort(self, *args): pass

class IFileRegistry:
    """
    provides for resource protection from
    concurrent thread access within a transaction
    scope
    """

class FileRegistry:

    __implements__ = (IFileRegistry,)

    def __init__(self):
        
        self._files = {}


    def add_file(self, file):

        try:
            file_registry_lock.acquire()

            if self._files.has_key(file):
                self.registry_collision(file)
            else:
                self._files[f]=None
        finally:
            file_registry_lock.release()
            
    def add_files(self, files):
        
        if not isinstance(file, ListType) or isinstance(file,TupleType):
            raise  TxnFsError("add files arg must be sequence")

        try:
            
            file_registry_lock.acquire()

            keys = self._files.keys()
            
            for f in files:
                if f in keys:
                    self.registry_collision(f)
                else:
                    self._files[f]=None
        finally:
            file_registry_lock.release()
            
    def has_file(self, path):
        return self._files.has_key(path)

    def registry_collision(self, path):
        
        # should probably raise a conflict error
        # so we can get retried
        
        raise TxnFsError("File is Locked")
    
    def remove_files(self, files):
        
        if not isinstance(file, ListType) or isinstance(file,TupleType):
            raise  TxnFsError("add files arg must be sequence")        

        try:
            file_registry_lock.acquire()
            
            for f in files:
                self._files[f]=None
        finally:
            file_registry_lock.release()

        return  1
        
file_registry_lock  =  RLock()
packagedir = join(gideon_home, 'var', 'packages')
tempdir    = join(gideon_home, 'var', 'tmp')
package_file_system =  PackageFileSystem(packagedir, tempdir)


random_notes = """

the filesystem needs to function as

transactionaly interface to heirarchical fs

fs interfaces
hash interface to fs -> package in subclass

encode  heirarchy info in temp store

place for process locks

a locked file registry for thread control over
   resource contention (files)
   
? make directories be locked as well

a lock file for process control
"""


from SimplePackageFs import *
